-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
-----
-- by @MegaKnight
--
-- lua tornadoes based on wind and particle emitters
--
-- DO NOT REUSE CODE WITHOUT PERMISSION
-----

local M = {}


-- variables

local nadoActive = false --> is active ?

local nadoCoords = {} --> coordinates of tornado, z all the way
nadoCoords.x = nil
nadoCoords.y = nil

local nadoDuration = 20 --> duration of tornado in seconds
--local nadoTimer = 0 --> tornado logic executed each 0.1 seconds --> NO, weird

local windSpeed = {} --> wind speed of tornado
windSpeed.x = 0
windSpeed.y = 0
windSpeed.z = 0 
local maxUpDrag = 95 --> maximum speed of upwards facing wind  

-- rotation
local yMax = true
local yMin = false
local xMax = false
local xMin = true
local rotation = true --> direction of rotation, true = clockwise

-- domain of influence, radius
local nadoRadius = 20 --> meters

-- dust particle emitter
local dustEmitter = nil

-- move tornado position every 3 seconds
local moveTimer = 0



-- initialize tornado
local function startTornado()

	-- don't do anything when already active tornado
	if nadoActive then return end
	
	-- set coordinates, random offset from player location
	local playerCoords = be:getPlayerVehicle(0):getPosition()
	nadoCoords.x = playerCoords.x + math.random(-30,30)
	nadoCoords.y = playerCoords.y + math.random(-30,30)
	
	-- get terrain height at (x,y), asssume terrain above z = 2000
	local zCoord = be:getSurfaceHeightBelow(vec3(nadoCoords.x, nadoCoords.y, 2000))
	
	-- set duration in seconds
	nadoDuration = math.random(60,300) --> random duration between 1 and 5 minutes
	
	-- reset variables
	--nadoTimer = 0
	moveTimer = 0
	windSpeed.x = 0
	windSpeed.y = 0
	windSpeed.z = 0
	
	-- set rotation
	yMax = true
	yMin = false
	xMax = false
	xMin = true
	
	-- random strength between 85 and 105, radius dependent on that
	maxUpDrag = math.random(85, 105)
	nadoRadius = maxUpDrag / 4.75 --> 95 => 20, standard tornado
	
	-- activate
	nadoActive = true
	
	-- place and enable particles
	--dustEmitter.position = Point3F(nadoCoords.x, nadoCoords.y, playerCoords.z) --> error
		--> use getColumn(col) to get column of matrix as vec3/table --> setColumn, cf. object.lua in ge/extensions/editor/api
	-- create table for matrix creation
	--local tbl = {}
	--tbl.c0 = {nadoCoords.x, 0, 0}
	--tbl.c1 = {0, nadoCoords.y, 0}
	--tbl.c2 = {0, 0, playerCoords.z}
	--tbl.c3 = {0, 0, 0}
	-- create matrix for point
	--local mtx = MatrixF(true)
	--mtx:setColumn(0, tbl.c0)
	--mtx:setColumn(1, tbl.c1)
	--mtx:setColumn(2, tbl.c2)
	--mtx:setColumn(3, tbl.c3)
	dustEmitter:setPosition(vec3(nadoCoords.x, nadoCoords.y, zCoord - 2)) -- FINALLY! YEEEEEEEEEEEEEEEEEEEEEEEEES!!!!
	dustEmitter.emitter = scenetree.findObject("BNGP_utah_dust_huge") --> is nil first time ???
	dustEmitter.active = true
	scenetree.findObjectById(dustEmitter:getId()):postApply()
	
	-- log nado pos, duration, radius and strength
	log("I", "", "tornado spawned: pos(" .. string.format("%.2f",nadoCoords.x) .. "," .. string.format("%.2f",nadoCoords.y) .. "), duration = " .. nadoDuration .. ", radius = " .. string.format("%.2f",nadoRadius) .. ", strength = " .. maxUpDrag)
	
	-- display message
	guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "A tornado is forming!", config = {timeOut = 5000}})
	
end



-- core tornado logic
local function doTornadoStep()
	-- do wind force logic
	-- rotate
	if yMax then
		if windSpeed.y > -50 then
			windSpeed.y = windSpeed.y - 5
		else
			windSpeed.y = windSpeed.y + 5
			yMin = true
			yMax = false
		end
	elseif yMin then
		if windSpeed.y < 50 then
			windSpeed.y = windSpeed.y + 5
		else
			windSpeed.y = windSpeed.y - 5
			yMax = true
			yMin = false
		end
	end
	if xMax then
		if windSpeed.x > -50 then
			windSpeed.x = windSpeed.x - 5
		else
			windSpeed.x = windSpeed.x + 5
			xMin = true
			xMax = false
		end
	elseif xMin then
		if windSpeed.x < 50 then
			windSpeed.x = windSpeed.x + 5
		else
			windSpeed.x = windSpeed.x - 5
			xMax = true
			xMin = false
		end
	end
	-- pull up
	if windSpeed.z < maxUpDrag then
		windSpeed.z = windSpeed.z + 1
	end
	
	-- check vehicles in domain of influence and apply force
	for i = 0, be:getObjectCount() do 
		local veh = be:getObject(i)
		if veh then --> does veh exist?
			local vehPos = veh:getPosition()
			local distX = vehPos.x - nadoCoords.x
			local distY = vehPos.y - nadoCoords.y --> calculate distance
			local dist = math.sqrt((distX * distX) + (distY * distY))
			if dist < nadoRadius then --> is veh inside domain of influence?
				veh:queueLuaCommand('obj:setWind(' .. windSpeed.x .. ', ' .. windSpeed.y .. ', ' .. windSpeed.z .. ')')
			end
		end
	end
	
end


-- moves tornado towards player
local function moveTornado()
	local playerCoords = getPlayerVehicle(0):getPosition()
	-- move tornado towards player
	if nadoCoords.x > playerCoords.x then
		nadoCoords.x = nadoCoords.x - 2
	else
		nadoCoords.x = nadoCoords.x + 2
	end
	if nadoCoords.y > playerCoords.y then
		nadoCoords.y = nadoCoords.y - 2
	else
		nadoCoords.y = nadoCoords.y + 2
	end
	-- get terrain height at (x,y), asssume terrain above z = 2000
	local zCoord = be:getSurfaceHeightBelow(vec3(nadoCoords.x, nadoCoords.y, 2000))
	-- move particle emitter
	dustEmitter:setPosition(vec3(nadoCoords.x, nadoCoords.y, zCoord - 1))
	scenetree.findObjectById(dustEmitter:getId()):postApply()
end


-- removes tornado
local function removeTornado()
	-- setwind force to 0
	log("I", "", "removed tornado.")
	be:queueAllObjectLua('obj:setWind(0, 0, 0)')
	
	nadoActive = false
	
	-- hide particles
	dustEmitter.active = false
	scenetree.findObjectById(dustEmitter:getId()):postApply()
end


-- creates tornadoes randomly
local function randomTornado()
	local randomVal = math.random()
	if randomVal > 0.95 then
		
		-- start tornado
		startTornado()
	end
end



-- executed each GFX frame
local doRandomTornadoes = false --> do random tornadoes ?
local randomTornadoTimer = 0
local function onUpdate(dt)

	if not simTimeAuthority.getPause() then
	local timerFactor = simTimeAuthority.getReal()

	-- do tornado logic only when tornado is active
	if nadoActive then
	
		--nadoTimer = nadoTimer + (dt * timerFactor)
		moveTimer = moveTimer + (dt * timerFactor)
		nadoDuration = nadoDuration - (dt * timerFactor)
		
		--if nadoTimer > 0.1 then --> execute tornado logic each 0.1 seconds --> not needed, is weird
		doTornadoStep() 
			--nadoTimer = 0
		--end
		
		if moveTimer > 3 then
			moveTimer = 0
			moveTornado()
		end
		
		if nadoDuration < 0 then
			nadoActive = false --> if timer runs out, deactivate and remove tornado
			removeTornado()
		end
	
	end
	
	-- random tornado, for testing
	if doRandomTornadoes and (not nadoActive) then 
		randomTornadoTimer = randomTornadoTimer + (1 * timerFactor)
		if randomTornadoTimer > 200 then
			randomTornado() 
			randomTornadoTimer = 0
		end
	end
	
	end
end


local function onExtensionLoaded()
	log("I", "", "Tornado extension loaded")
	
	-- does particle emitter exist?
	dustEmitter = scenetree.findObject("tornado_dust_MKDW")
	
	if not dustEmitter then
		-- if not, create particle emitter
		dustEmitter = createObject("ParticleEmitterNode")
		dustEmitter.scale = Point3F(1, 1, 1)
		--dustEmitter.position = Point3F(1, 1, 1) --> error, seems to be matrix instead of Point3F/table
		dustEmitter.emitter = scenetree.findObject("BNGP_utah_dust_huge")
		dustEmitter.dataBlock = scenetree.findObject("lightExampleEmitterNodeData1")
		dustEmitter.velocity = 1.0
		dustEmitter.active = false
		dustEmitter.hidden = false
		dustEmitter:registerObject('tornado_dust_MKDW')
	end
	
end

local function onExtensionUnloaded()
	log("I", "", "Tornado extension unloaded")
end


--local function onClientEndMission()
	-- cleanDisable() called by weather extension
--end


-- called by dynamic weather, prevent crashes
local function cleanDisable()
	-- remove wind
	be:queueAllObjectLua('obj:setWind(0, 0, 0)')
	
	-- hide particles
	if scenetree.findObjectById(dustEmitter:getId()) then
		scenetree.findObjectById(dustEmitter:getId()):delete()
	end
	
	-- unload
	extensions.unload('dynamicWeatherMK_tornado')
end



-----
-- TODO:
-- - radius of influence   				 (DONE)
-- - tornado movement					 (DONE)
-- - particles or something 			 (DONE)
-- - random rotation, speed and stuff
-- - make z coordinate dependent on terrain/objects (DONE)
-- - make multiple tornadoes possible ? --> seperate mod
-----


-- public interface

M.onUpdate = onUpdate --> executed by game engine
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
--M.onClientEndMission = onClientEndMission
M.cleanDisable = cleanDisable

M.startTornado = startTornado --> console interaction
M.toggleRandomTornadoes = function() 
							doRandomTornadoes = not doRandomTornadoes 
							if doRandomTornadoes then 
								log("I", "", "Random tornadoes turned on.") 
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Random tornado spawning turned on.", config = {timeOut = 5000}})
							else
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Random tornado spawning turned off.", config = {timeOut = 5000}})
								log("I", "", "Random tornadoes turned off.")
							end
						  end
M.removeTornado = removeTornado


return M



